#!/usr/bin/perl
#
# Copyright © 2008,2010,2012,2019 Nick Bowler
#
# Silly little script to generate an FVWM menu with various bits of MPD
# status information and controls.
#
# License GPLv3+: GNU General Public License version 3 or any later version.
# This is free software: you are free to change and redistribute it.
# There is NO WARRANTY, to the extent permitted by law.

use strict;

use utf8;

use Encode qw(decode encode);
use Encode::Locale qw(decode_argv);
decode_argv(Encode::FB_CROAK);
binmode(STDOUT, ":utf8");

use IO::Socket::INET6;
use Getopt::Long qw(:config gnu_getopt);
use Scalar::Util qw(reftype);
use List::Util qw(any);
use FindBin;

use constant {
	MPD_MJR_MIN => 0,
	MPD_MNR_MIN => 21,
	MPD_REV_MIN => 0,
};

my $SELF = "$FindBin::Bin/$FindBin::Script";

my $MUSIC = $ENV{MUSIC}    // "/srv/music";
my $host  = $ENV{MPD_HOST} // "localhost";
my $port  = $ENV{MPD_PORT} // "6600";
my $sock;

my $menu;
my $mode = "top";

# Quotes the argument so that it is presented as a single argument to MPD
# at the protocol level.  This also works OK for most FVWM arguments.
sub escape {
	my $s = @_[0] // $_;

	# No way to encode literal newlines in the protocol, so we
	# convert any newlines in the arguments into a space, which
	# can help with quoting.
	$s =~ s/\n/ /g;

	if (/[ \t\\"]/) {
		$s =~ s/[\\"]/\\$&/g;
		return "\"$s\"";
	}

	$s =~ s/^\s*$/"$&"/;
	return $s;
}

# Submit a command to the MPD server; each argument to this function
# is quoted and sent as a single argument to MPD.
sub mpd_exec {
	my $cmd = join(' ', map { escape } @_);

	print $sock "$cmd\n";
}

sub fvwm_cmd_unquoted {
	print join(' ', @_), "\n";
}

sub fvwm_cmd {
	fvwm_cmd_unquoted(map { escape } @_);
}

# Quotes the argument in such a way that it is passed unadulterated by
# both FVWM and the shell to a command as a single argument (for use as
# an # argument for e.g., the Exec or PipeRead FVWM commands).
#
# The result must be used with fvwm_cmd_unquoted;
sub fvwm_shell_literal {
	my $s = @_[0] // $_;

	$s =~ s/\$/\$\$/g;
	if ($s =~ /[' \t]/) {
		$s =~ s/'/'\\''/g;
		return "'$s'";
	}
	$s =~ s/^\s*$/'$&'/;
	return "$s";
}

# Escapes metacharacters in the argument used in FVWM menu labels.  The
# string must still be quoted (e.g., by using fvwm_cmd).
sub fvwm_label_escape {
	my @tokens = split /\t/, $_[0];
	@tokens[0] =~ s/&/&&/g;
	my $ret = join "\t", @tokens;
	$ret =~ s/[\$@%^*]/$&$&/g;
	return $ret;
}

# make_submenu(name, [args ...])
#
# Creates a submenu (with the specified name) constructed by invoking this
# script with the given arguments.  Returns a list that can be passed to
# fvwm_cmd to display the menu.
sub make_submenu {
	my $name = shift;
	$name =~ s/-/_/g;
	unshift @_, ("exec", $SELF, "--menu=$name");

        fvwm_cmd("DestroyFunc", "Make$name");
        fvwm_cmd("AddToFunc", "Make$name");
        fvwm_cmd("+", "I", "DestroyMenu", $name);

        fvwm_cmd("DestroyMenu", $name);
        fvwm_cmd("AddToMenu", $name, "DynamicPopupAction", "Make$name");
        fvwm_cmd("AddToFunc", "KillMenuMPD", "I", "DestroyMenu", $name);

        fvwm_cmd("DestroyFunc", "Make$name");
        fvwm_cmd("AddToFunc", "Make$name");
        fvwm_cmd("+", "I", "DestroyMenu", $name);
        fvwm_cmd("+", "I", "-PipeRead",
	         join(' ', map { fvwm_shell_literal } @_));
        fvwm_cmd("AddToFunc", "KillMenuMPD", "I", "DestroyFunc", "Make$name");

	return ("Popup", $name);
}

# get_item_thumbnails({ options }, file, ...)
# get_item_thumbnails(file, ...)
#
# For each music file listed, obtain a thumbnail (if any) for the cover art.
#
# The first argument is a hash reference to control the mode of operation;
# it may be omitted for default options.
#
#   get_item_thumbnails({ small => 1 }, ...) - smaller thumbnails
#
# The returned list consists of strings (in the same order as the filename
# arguments) suitable for use directly in FVWM menus; by default the filename
# is bracketed by asterisks (e.g., "*thumbnail.png*"); in small mode it is
# surrounded by % (e.g., "%thumbnail.png%").  If no cover art was found, the
# empty string is returned for that file.
sub get_item_thumbnails {
	my @results = ();
	my $flags = {};
	my @opts = ();

	$flags = shift if (reftype($_[0]) eq "HASH");
	return @results unless @_;

	my $c = "*";
	if ($flags->{small}) {
		push @opts, "--small";
		$c = "%";
	}

	open THUMB, "-|", "$FindBin::Bin/mpdthumb.sh", @opts, "--", @_;
	foreach (@_) {
		my $thumb = <THUMB>;
		chomp $thumb;

		$thumb = "$c$thumb$c" if (-f $thumb);
		push @results, $thumb;
	}
	close THUMB;
	die("mpdthumb failed") if ($?);

	return @results;
}

# add_track_metadata(hashref, key, value)
#
# Inserts the given key into the referenced hash; if the key already exists
# in the hash then the hash element is converted to an array reference (if
# it isn't already) and the value is appended to that array.
sub add_track_metadata {
	my ($entry, $key, $value) = @_;

	if (exists($entry->{$key})) {
		my $ref = $entry->{$key};

		if (reftype($ref) ne "ARRAY") {
			return if ($ref eq $value);

			$ref = [$ref];
			$entry->{$key} = $ref;
		}

		push(@$ref, $value) unless (any {$_ eq $value} @$ref);
	} else {
		$entry->{$key} = $value;
	}
}

# get_track_metadata(hashref, key)
#
# Return the values associated with the given metadata key as a list.
sub get_track_metadata {
	my ($entry, $key) = @_;

	return () unless (exists($entry->{$key}));

	my $ref = $entry->{$key};
	return @$ref if (reftype($ref) eq "ARRAY");
	return $ref
}

# Given a music filename, search for the cover art in the same directory.
sub mpd_cover_filename {
	my ($dir) = @_;
	my $file;

	$dir =~ s/\/[^\/]*$//;
	foreach ("cover.png", "cover.jpg", "cover.tiff", "cover.bmp") {
		if (-f "$dir/$_") {
			$file = "$dir/$_";
			last;
		}
	}
	return unless defined $file;

	# Follow one level of symbolic link to get to the scans directory.
	$file = readlink($file) // $file;
	$file = "$dir/$file" unless ($file =~ /^\//);
	return $file;
}

# Generate the cover art entry in the top menu.
sub top_track_cover {
	my ($entry) = @_;

	($entry->{thumb}) = get_item_thumbnails($entry->{file});
	print "$entry->{thumb}\n";
	if ($entry->{thumb}) {
		my $file = "$MUSIC/$entry->{file}";
		my $cover = mpd_cover_filename($file);

		$cover = fvwm_shell_literal($cover // $file);
		fvwm_cmd_unquoted("AddToMenu", escape($menu),
		                  escape($entry->{thumb}),
		                  "Exec", "exec", "geeqie", $cover);
	}
}

# Generate the "Title:" entry in the top menu.
sub top_track_title {
	my ($entry) = @_;

	my @submenu = make_submenu("$menu-TopTrack",
	                           "--title=$entry->{Title}");

	fvwm_cmd("AddToMenu", $menu,
	         fvwm_label_escape("Title:\t$entry->{Title}"),
	         @submenu);
}

# Generate the "Artist:" entry in the top menu.
sub top_track_artist {
	my ($entry) = @_;

	my @submenu = make_submenu("$menu-TopArtist",
	                           "--artist=$entry->{Artist}");

	fvwm_cmd("AddToMenu", $menu,
	         fvwm_label_escape("Artist:\t$entry->{Artist}"),
	         @submenu);
}

# Generate the "Album:" entry in the top menu.
sub top_track_album {
	my ($entry) = @_;
	my @submenu;

	my @submenu = make_submenu("$menu-TopAlbum",
	                           "--artist=$entry->{Artist}",
				   "--album=$entry->{Album}");

	fvwm_cmd("AddToMenu", $menu,
	         fvwm_label_escape("Album:\t$entry->{Album}"),
	         @submenu);
}

# Global hash for tracking what is to be "accepted".
my %accept = ();

# Default values for stuff.
my ($album, $artist, $title);

sub print_version {
	print <<EOF
mpdmenu.pl 0.8
Copyright © 2019 Nick Bowler
License GPLv3+: GNU General Public License version 3 or any later version.
This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law.
EOF
}

sub print_usage {
	my $fh = $_[1] // *STDERR;

	print $fh "Usage: $0 [options]\n";
	print "Try $0 --help for more information.\n" unless (@_ > 0);
}

sub print_help {
	print_usage(*STDOUT);
	print <<EOF
This is "mpdmenu": a menu-based MPD client for FVWM.

Options:
  -h, --host=HOST   Connect to the MPD server on HOST, overriding defaults.
  -p, --port=PORT   Connect to the MPD server on PORT, overriding defaults.
  -m, --menu=NAME   Set the name of the generated menu.
  -V, --version     Print a version message and then exit.
  -H, --help        Print this message and then exit.
EOF
}

GetOptions(
	'host|h=s'   => \$host,
	'port|p=s'   => \$port,
	'menu|m=s'   => \$menu,

	'album=s'    => sub { $album = $_[1]; $mode = "album"; },
	'artist=s'   => sub { $artist = $_[1];
	                      $mode = "artist" unless $mode eq "album"; },
	'title=s'    => sub { $title = $_[1]; $mode = "track"; },

	'V|version'  => sub { print_version(); exit },
	'H|help'     => sub { print_help(); exit },
) or do { print_usage; exit 1 };

# Connect to MPD.
$sock = new IO::Socket::INET6(
	PeerAddr => $host,
	PeerPort => $port,
	Proto => 'tcp',
	Timeout => 2
) or die("could not open socket: $!.\n");
binmode($sock, ":utf8");

die("could not connect to MPD: $!.\n")
	if (!(<$sock> =~ /^OK MPD ([0-9]+)\.([0-9]+)\.([0-9]+)$/));

die("MPD version $1.$2.$3 insufficient.\n")
	if (  ($1 <  MPD_MJR_MIN)
	   || ($1 == MPD_MJR_MIN && $2 <  MPD_MNR_MIN)
	   || ($1 == MPD_MJR_MIN && $2 == MPD_MNR_MIN && $3 < MPD_REV_MIN));

if ($mode eq "top") {
	my %current;
	my %state;

	$menu //= "MenuMPD";

	mpd_exec("status");
	while (<$sock>) {
		last if (/^OK/);
		die($_) if (/^ACK/);

		if (/^(\w+): (.*)$/) {
			$state{$1} = $2;
		}
	}

	mpd_exec("currentsong");
	while (<$sock>) {
		last if (/^OK/);
		die($_) if (/^ACK/);

		if (/^(\w+): (.*)$/) {
			add_track_metadata(\%current, $1, $2);
		}
	}

	my $playstate = $state{state} eq "play"  ? "Playing"
	              : $state{state} eq "stop"  ? "Stopped"
	              : $state{state} eq "pause" ? "Paused"
	              : "Unknown";
	fvwm_cmd("AddToMenu", $menu, $playstate, "Title");

	if (exists($current{file})) {
		top_track_cover(\%current);
		top_track_title(\%current);
		top_track_artist(\%current);
		top_track_album(\%current);
	} else {
		fvwm_cmd("AddToMenu", $menu, "[current track unavailable]");
	}

	if ($state{state} =~ /^p/) {
		my $pp = $state{state} eq "pause" ? "lay" : "ause";

		fvwm_cmd("AddToMenu", $menu, "", "Nop");
		fvwm_cmd("AddToMenu", $menu, "Next%next.svg:16x16%",
		       "Exec", "exec", "$FindBin::Bin/mpdexec.pl", "next");
		fvwm_cmd("AddToMenu", $menu, "P$pp%p$pp.svg:16x16%",
		       "Exec", "exec", "$FindBin::Bin/mpdexec.pl", "p$pp");
		fvwm_cmd("AddToMenu", $menu, "Stop%stop.svg:16x16%",
		       "Exec", "exec", "$FindBin::Bin/mpdexec.pl", "stop");
		fvwm_cmd("AddToMenu", $menu, "Prev%prev.svg:16x16%",
		       "Exec", "exec", "$FindBin::Bin/mpdexec.pl", "previous");
	} elsif ($state{state} eq "stop") {
		fvwm_cmd("AddToMenu", $menu, "", "Nop");
		fvwm_cmd("AddToMenu", $menu, "Play%play.svg:16x16%",
		       "Exec", "exec", "$FindBin::Bin/mpdexec.pl", "play");
	}
} elsif ($mode eq "album") {
	# Create an album menu.
	my @playlist = ();
	my $entry;

	$menu = "MenuMPDAlbum" unless defined $menu;

	$album =~ s/"/\\"/g;
	print $sock "playlistfind album \"$album\"\n";
	while (<$sock>) {
		last if (/^OK/);
		die($_) if (/^ACK/);

		if (/^(\w+): (.*)$/) {
			if ($1 eq "file") {
				if (keys(%$entry) > 0) {
					addalbumentry(\@playlist, $entry)
				}

				$entry = {};
			}

			$entry->{$1} = $2;
		}
	}
	addalbumentry(\@playlist, $entry) if (keys(%$entry) > 0);

	die("No tracks found.\n") if (!@playlist);
	foreach (sort albumsort @playlist) {
		my ($t_file, $t_trackno, $t_artist, $t_title, $t_id) = (
			$_->{file},
			$_->{Track},
			$_->{Artist},
			$_->{Title},
			$_->{Id},
		);

		next if (defined $artist && !$accept{albumdir($t_file)});

		$t_artist = sanitise($t_artist, 0);
		$t_title  = sanitise($t_title, 0);

		my $cmd = sprintf "AddToMenu $menu \"%d\t%s - %s\""
		                  ." Exec exec $FindBin::Bin/mpdexec.pl"
				  ." playid %d",
		                  $t_trackno, $t_artist, $t_title, $t_id;

		cmd($cmd);
	}
} elsif ($mode eq "artist") {
	# Create an artist menu.
	my %albums = ();
	my $file;
	my $quoteartist = $artist;

	$menu = "MenuMPDArtist" unless defined $menu;

	$quoteartist =~ s/"/\\"/g;
	print $sock "playlistfind artist \"$quoteartist\"\n";
	while (<$sock>) {
		last if (/^OK/);
		die($_) if (/^ACK/);

		if (/^(\w+): (.*)$/) {
			$file       = $2    if ($1 eq "file");
			$albums{$2} = $file if ($1 eq "Album");
		}
	}

	die("No albums found.\n") if (!keys(%albums));

{ # work around 'use locale' breaking s///i
	my $i = 0;
	use locale;

	my @keys = sort keys %albums;
	my @thumbs = get_item_thumbnails({ small => 1 },
	                                  map { $albums{$_} } @keys);

	foreach my $key (@keys) {
		my $a_album  = sanitise($key, 1);
		my $thumb = shift @thumbs;

		cmd("AddToMenu $menu \"$thumb$a_album\" Popup MenuMPDArt_$i");

		cmd("AddToMenu MenuMPDArt_$i DynamicPopUpAction MakeMenuMPDArt_$i");

		cmd("DestroyFunc MakeMenuMPDArt_$i");
		cmd("AddToFunc   MakeMenuMPDArt_$i
		     + I DestroyMenu MenuMPDArt_$i
		     + I -PipeRead \"exec $SELF "
		           ."--menu MenuMPDArt_$i "
		           ."--album  ".shellify($key, 1)." "
		           ."--artist ".shellify($artist, 1)."\"");

		cmd("AddToFunc KillMenuMPD I DestroyMenu MenuMPDArt_$i");
		cmd("AddToFunc KillMenuMPD I DestroyFunc MakeMenuMPDArt_$i");

		$i++;
	}
} # end use locale workaround
} elsif ($mode eq "track") {
	# Create a title menu.
	my @titles;
	my $entry;

	$menu = "MenuMPDTitle" unless defined $menu;

	# Open and close brackets.
	my ($ob, $cb) = ("[\[~〜<〈(ー−-]", "[\]~〜>〉)ー−-]");

	$_ = $title;

	# Deal with specific cases.
	s/ちいさな(?=ヘミソフィア)//;                 # ヘミソフィア
	s/ "mix on air flavor" dear EIKO SHIMAMIYA//; # Spiral wind
	s/ "So,you need me" Style//;                  # I need you
	s/ ::Symphony Second movement:://;            # Disintegration
	s/-\[instrumental\]//;                        # 青い果実
	s/ -Practice Track-//;                        # Fair Heaven
	s/〜世界で一番アナタが好き〜//;               # Pure Heart
	s/〜彼方への哀歌//;                           # 十二幻夢
	s/　sora no uta ver.//;                       # 美しい星

	s/\s*-remix-$//; # Otherwise "D-THREAD -remix-" doesn't work right.

	# Deal with titles like "blah (ABC version)".
	s/\s*$ob.*(style|mix|edit|edition|ver\.?|version|melody|カラオケ)$cb?$//i;

	# Deal with titles like "blah (without XYZ)".
	s/\s*$ob\s*((e\.)?piano|english|japanese|inst|tv|without|w\/o|off|back|short|karaoke|game).*//i;

	# Deal with titles like "blah instrumental".
	s/\s+(instrumental|off vocal|short|tv)([\s-]+(mix|size|version))?$//i;
	s/\s+without\s+\w+$//i;

	# Deal with separate movements in classical pieces.
	s/: [IVX]+\..*//;

	my $basetitle  = $_;
	my $_basetitle = $basetitle;

	$_basetitle =~ s/"/\\"/g;
	print $sock "playlistsearch title \"$_basetitle\"\n";
	while (<$sock>) {
		last if (/^OK/);
		die($_) if (/^ACK/);

		if (/^(\w+): (.*)$/) {
			if ($1 eq "file") {
				push @titles, $entry if (keys(%$entry) > 0);
				$entry = {};
			}

			$entry->{$1} = $2;
		}
	}
	push @titles, $entry if (keys(%$entry) > 0);

{ # work around 'use locale' breaking s///i
	use locale;

	my @thumbs = get_item_thumbnails({ small => 1 },
	                                  map { $_->{file} } @titles);
	for (my $i = 0; $i < @titles; $i++) {
		$titles[$i]->{thumb} = $thumbs[$i];
	}

	foreach (sort titlesort @titles) {
		my ($t_file, $t_artist, $t_title, $t_id, $thumb) = (
			$_->{file},
			$_->{Artist},
			$_->{Title},
			$_->{Id},
			$_->{thumb}
		);

		# MPD searches are case-insensitive.
		next if (!($t_title =~ m/(\P{Latin}|^)\Q$basetitle\E(\P{Latin}|$)/ || $t_title =~ m/\Q$basetitle\E/i));

		$t_artist = sanitise($t_artist, 1);
		$t_title  = sanitise($t_title, 1);

		cmd("AddToMenu $menu \"$thumb$t_artist - $t_title\""
		    ." Exec exec $FindBin::Bin/mpdexec.pl"
		    ." playid $t_id");
	}
} # end use locale workaround
}

# Finished.
print $sock "close\n";

sub sanitise
{
	$_ = $_[0];
	s/&/&&/g if ($_[1]);
	s/([\$@%^*])/\1\1/g;
	s/"/\\"/g;
	return $_;
}

sub addalbumentry
{
	my ($playlist, $entry) = @_;

	push(@$playlist, $entry);

	if (defined $artist && $artist eq $entry->{Artist}) {
		my $albumdir = albumdir($entry->{file});
		$accept{$albumdir} = 1;
	}
}

sub albumdir
{
	my $file = $_[0];

	$file =~ s:(/Disk [0-9]+[^/]*)?/[^/]*$::;
	return $file
}

sub albumsort
{
	return ($a->{Disc} <=> $b->{Disc}) if ($a->{Disc} != $b->{Disc});
	return ($a->{Track} <=> $b->{Track});
}

sub titlesort
{
	return ($a->{Album}  cmp $b->{Album})  if($a->{Album}  ne $b->{Album});
	return ($a->{Artist} cmp $b->{Artist}) if($a->{Artist} ne $b->{Artist});
	return ($a->{Title}  cmp $b->{Title});
}

sub shellify
{
	my ($str, $quoted) = @_;
	$str =~ s/'/'\\''/g;
	if ($quoted) {
		$str =~ s/\\/\\\\/g;
		$str =~ s/"/\\"/g;
	}
	return "'$str'";
}

sub cmd
{
	print "$_[0]\n";
}
